// Main.c - Main Code For GBA Demo
// ---------------------------------------------------
// Copyright (C)Savage Development.
// By Matthew Tighe

#include "gba.h"
#include "fonts\font.h"		// test tile data
#include "gfx\logo.h"		// Savage logo
#include "gfx\mask.h"

// font structure, describes a font in character vram
struct font
{
	u16	dwStart;			// Starting char num in vram
	s16 dwAdjust;			// Adjustment for chars not in font
	u16 dwHeight;			// Height of font in chars
	u16 dwWidth;			// Width of font in chars
};

typedef struct font FONT;

// Memory / String Handling Functions
// ----------------------------------
void bytecpy( u8 *pDest, u8 *pSource, int nCount );
int strLen(char *szText);

// Map Handling Functions
// ----------------------
void mapset( u16 *pDest, u16 dwSource, int nCount );
void mapPrint( int x, int y, u16 *pMap, char *szText, FONT *pFont);
void mapSet( int x, int y, int width, int height, u16 *pDest, u16 tile);
void mapPut( int x, int y, int width, int height, u16 *pDest, u16 *pSrc, u16 start);
void mapPutLinear( int x, int y, int width, int height, u16 *pDest, u16 start);

// Screen Handling Functions
// -------------------------
void scrWaitVbl();
void scrWaitVSync();
u16 scrIsVbl();
void palCopy( u16 *pPal, int nStart, u16 *pColours, int nCount );
void sfxBgMosaic(int wsize, int hsize);
void sfxObjMosaic(int wsize, int hsize);

// Hardware Sprite Handling Functions
// ----------------------------------

void sprClear( int nStart, int nCount );


int MakeTextSprites(int x, int y, char *szText, FONT *pFont);


char text[10][32] = {"Welcome to the SavageGBA demo",
					"By Matthew Tighe",
					"It some blending with sprites",
					"Check out my website at....",
					"www.matt-tighe.co.uk/gba",
					"for GBA dev-tools,docs,demos",
					"Other Great Websites",
					"Jeff Frohwein's at www.devrs.com",
					"GBADEV at www.gbadev.org",
					"Thanks for watchin!!!!!" };

void AgbMain(void)
   {
    u16	*pFgMap;					// Pointer to Mask map
	u16 *pBgMap;					// Pointer to Savage map
	u8 *pTiles;						// Pointer to tiles
	u8 *pObjTiles;
	FONT fntSystem;
	OAMOBJECT *pSprites;
	int alpha, alphainc,i, maskxinc, maskyinc;
	int textlen, textx,textaction,textycount, textcount = 0, textwait = 0;


	pFgMap = (u16*)0x6000000;
	pBgMap = (u16*)0x6003000;

	pTiles = (u8*)0x6004000;
	pObjTiles = (u8*)0x6010000;
	pSprites = ( OAMOBJECT * ) 0x07000000;

	DISPCNT = DISP_MODE(0) | DISP_BG0 | DISP_BG1 | DISP_OBJ;				// Set Mode 0, BG0 on
	BG0CNT = BG_CHARBASE(0x4000) | BG_COLOUR_256x1 | BG_T512x512;			// BG 0 Characters at 0x4000
	BG0X = 16;
	BG0Y = 48;

	BG1CNT = BG_CHARBASE(0x4000) | BG_MAPBASE(0x3000) | BG_COLOUR_256x1 | BG_T256x256;			// BG 0 Characters at 0x4000
	BG1X = 0;
	BG1Y = 0;

	// Setup initial scene, with logo blown-up and mask faded.
	
	BLDMOD = BLEND_ALPHA | BLEND_SRC(1) | BLEND_DEST(2);
	COLEV = 0;

	BG1CNT = BG1CNT | BG_MOSAIC;
	sfxBgMosaic( 15, 15 );

	// Load the video characters and palettes

	sprClear(0, 128 );
	palCopy( (u16 *) BG_PAL, 0, logoPal, logoPalLen );
	palCopy( (u16 *) OBJ_PAL, 0, fontPal, fontPalLen );
	
	bytecpy( pTiles, logoData, logoDataLen);					// Copy logo characters
	bytecpy( (pTiles + logoDataLen), maskData, maskDataLen);			// Copy mask characters
	bytecpy( pObjTiles, fontData, 7744);						// Copy font characters
	
	// Set up font descriptor

	fntSystem.dwStart = 1;
	fntSystem.dwAdjust = -32;
	fntSystem.dwHeight = 1;
	fntSystem.dwWidth = 1;

	// Set up the maps

	mapSet( 0,0,64,64, pFgMap, logoTiles+1 );
	mapPut( 0,2, 30,16, pBgMap, logoMap,0);
	mapPut(5,13, 25,7, pFgMap, maskMap,logoTiles);
	
	// Un-mosaic the logo
	
	for(i = 15; i > 0; i--)
	{
		sfxBgMosaic(i,i);
		scrWaitVSync();
		scrWaitVSync();
	}
	BG1CNT = BG1CNT ^ BG_MOSAIC;		// Turn off bk mosaic

	textaction = 0;
	textlen = 0;
	textycount = 0;
	textx = 0;

	alpha = 0;
	alphainc = 1;
	maskyinc = -1;
	maskxinc = 1;
	do
	{

		switch( textaction )
		{
			case 0:		// Loading new text
				textlen = MakeTextSprites(240, 76, text[textcount], &fntSystem);
				textcount++;
				if(textcount >= 8)
					textcount = 0;
				textaction = 1;							// 1 means comming onto the screen
				textx = ( 240 - (textlen << 3) ) >> 1;	// Center horizontally
				textycount = 0;
				textwait = 0;

				break;
			case 1:		// Scrolling onto screen from right
				for( i = 0; i < textlen; i++)		// For each sprite character
					if( (pSprites[i].x - ( i << 3 ) > textx ) )
						pSprites[i].x -= 2;
					else
						textaction = 2;
				break;

			case 2:		// Waiting
				textwait++;
				if(textwait >= 120)
					textaction = 3;
				break;

			case 3:		// Scrolling off screen alternate top and bottom
				for( i = 0; i < textlen; i++)
				{
					if( textycount < 70 )
					{
						if( i & 1 )
							pSprites[i].y += 2;
						else
							pSprites[i].y -= 2;
					}
					else
						textaction = 0;

				}
				textycount++;

				break;
		}

		// Flash the logo using alpha fading
		COLEV = 256*(16-(alpha>>1))+(alpha>>1);
		alpha += alphainc;

		if(alpha >= 16 || alpha <= 0)
			alphainc = (0 - alphainc);

		// Bounce the mask around the screen
		BG0X += maskxinc;

		if( BG0X >= 40 || BG0X <= 0 )		// change direction at edges.
			maskxinc = (0 - maskxinc);

		BG0Y += maskyinc;

		if( BG0Y >= 104 || BG0Y <= 0 )		// change direction at edges.
			maskyinc = (0 - maskyinc);

		scrWaitVSync();
	} while(1);
}
int MakeTextSprites(int x, int y, char *szText, FONT *pFont)
{	
	int i;
	OAMOBJECT *pSprites;

	pSprites = ( OAMOBJECT * ) 0x07000000;

	sprClear(0,128);

	for(i=0; i < strLen( szText ); i++)
	{
		pSprites[i].y = OBJ_COLOUR_256x1 | y;
		pSprites[i].x = (x + (i << 3) );
		pSprites[i].tile = ( u16 ) ( u8 ) ( (szText[i]  + pFont->dwAdjust ) << 1);
	}
	return(i);
}




void bytecpy( u8 *pDest, u8 *pSource, int nCount )
{
	int i;

	for( i = 0; i < nCount; i++ )
		pDest[i] = pSource[i];
}

void mapset( u16 *pDest, u16 dwSource, int nCount )
{
	int i;

	for( i = 0; i < nCount; i++ )
		pDest[i] = dwSource;
}

void mapPut( int x, int y, int width, int height, u16 *pDest, u16 *pSrc, u16 start)
{
	int i,j;
	u32 skew;
		
	pDest += ( (y * 32) + x );
	skew = ( 32 - width );

	for(j = 0; j < height; j++)
	{
		for(i = 0; i < width; i++)
			*pDest++ = (start + *pSrc++);

		pDest += skew;
	}
}

void mapSet( int x, int y, int width, int height, u16 *pDest, u16 tile)
{
	int i,j;
	u32 skew;
		
	pDest += ( (y * 32) + x );
	skew = ( 32 - width );

	for(j = 0; j < height; j++)
	{
		for(i = 0; i < width; i++)
			*pDest++ = tile;

		pDest += skew;
	}
}


void mapPutLinear( int x, int y, int width, int height, u16 *pDest, u16 start)
{
	int i,j;
	u32 skew;
		
	pDest += ( (y * 32) + x );
	skew = ( 32 - width );

	for(j = 0; j < height; j++)
	{
		for(i = 0; i < width; i++)
			*pDest++ = start++;

		pDest += skew;
	}
}


void mapPrint( int x, int y, u16 *pMap, char *szText, FONT *pFont)
{
	int i;
	u16 dwChar;

	i = 0;

	pMap += ( (y * 64) + x );

	do
	{
		dwChar = ( u8 ) ( szText[i] );

		pMap[i] = dwChar + pFont->dwAdjust;
		i++;

	} while( szText[i] != '\0' );
}

int strLen(char *szText)
{
	int i;
	i = 0;
	while( szText[i] != '\0' )
		i++;

	return( i );
}

// Wait until vertical blank
void scrWaitVbl()
{
	while( !scrIsVbl() );
}
// Wait for NEXT vertical blank
void scrWaitVSync()
{
	while( scrIsVbl() );
	while( !scrIsVbl() );
}

// Is the screen in verticle blank?
u16 scrIsVbl()
{
	return( DISPSTAT & STAT_VBL_MASK );
}

void palCopy( u16 *pPal, int nStart, u16 *pColours, int nCount )
{
	int i;

	for (i = nStart;i < nCount; i++)
	{
		pPal[i]=pColours[i];
	}
}

void sprClear( int nStart, int nCount )
{
	int i;
	OAMOBJECT *pSprites;

	pSprites = ( OAMOBJECT * ) 0x07000000;

	for(i = nStart; i < nCount; i++)
	{
		pSprites[i].x = 0;
		pSprites[i].y = 0;
		pSprites[i].tile = 0;
		pSprites[i].fx = 0;
	}
}

void sfxObjMosaic(int wsize, int hsize)
{
	MOSAIC &= ~0xFF00;
	MOSAIC |= ( ( ( (wsize & 15) << 4) | ( hsize & 15 ) )  << 8 );
}

void sfxBgMosaic(int wsize, int hsize)
{
	MOSAIC &= ~0x00FF;
	MOSAIC |= ( ( ( (wsize & 15) << 4) | ( hsize & 15 ) ) );
}